#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <stdbool.h>

#define DELKA_BUFFERU 16384
#define DELKA_CISLA 32
#define PRAH -1
#define SOUBOR_SETUP "setup.txt"
#define SOUBOR_DATA_TRAINING "DataTraining.txt"
#define SOUBOR_DATA_DEV_TEST "DataDevTest.txt"
#define SOUBOR_PERFORMANCE "performance.txt"
#define SOUBOR_VYSLEDEK "vysledek.txt"
#define SOUBOR_VYSLEDEK_DEV_TEST "vysledek_test.txt"
#define SOUBOR_VAHY "vahy.txt"
#define SOUBOR_WEIGHTS_BIN "weights.bin"
#define SOUBOR_NEJLEPSI_BIN "nejlepsi.bin"
#define real_rand() ((double) rand() / (RAND_MAX + 1.0) - 0.5)
//#define CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER
//#define CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER
#define CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
//#define CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER

typedef struct {
  char performance_tren[DELKA_CISLA];
  int pocet_spravnych_vzoru_tren;
  char performance_dev_test[DELKA_CISLA];
  int pocet_spravnych_vzoru_dev_test;
  int pocet_kroku;
} NEJLEPSI;

void ceske_cislo(char *s, double cislo)
{
  char *p_c;
  snprintf(s, DELKA_CISLA, "%.*f", DELKA_CISLA - 1, cislo);
  if ((p_c = strchr(s, '.')) != NULL)
    *p_c = ',';
}

unsigned cti_unsigned_int_ze_souboru(FILE *fr)
{
  int c;
  unsigned cislo;
  if (fscanf(fr, "%u", &cislo) != 1)
    printf ("Chybi hodnota v souboru!\n");
  printf("Precteno %u", cislo);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
  return cislo;
}

int cti_int_ze_souboru(FILE *fr)
{
  int c;
  int cislo;
  if (fscanf(fr, "%d", &cislo) != 1)
    printf ("Chybi hodnota v souboru!\n");
  printf("Precteno %d", cislo);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
  return cislo;
}

double cti_double_ze_souboru(FILE *fr)
{
  int c;
  double cislo;
  if (fscanf(fr, "%lf", &cislo) != 1)
    printf ("Chybi hodnota v souboru!\n");
  printf("Precteno %g", cislo);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
  return cislo;
}

void kopiruj_soubor(char *jmeno_souboru, FILE *fw)
{
  FILE *fr;
  int c;
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  while ((c = getc(fr)) != EOF)
    putc(c, fw);
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
}

int zjisti_pocet_radku_dat(char *jmeno_souboru)
{
  FILE *fr;
  char buffer[DELKA_BUFFERU];
  int pocet_radku = 0;
  double cislo;
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  while (fgets(buffer, DELKA_BUFFERU, fr)) {
    if (sscanf(buffer, "%lf", &cislo) == 1) {
      pocet_radku++;
    }
    buffer[0] = '\0';
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
  return pocet_radku;
}

void zaloz_matici_double(int radek, int sloupek, double ***x, double **p_x)
{
  int i;
  if ((*p_x = (double *) malloc(radek * sloupek * sizeof(double))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  if ((*x = (double **) malloc(radek * sizeof(double *))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  for (i = 0; i < radek; i++) {
    (*x)[i] = *p_x + i * sloupek;
  }
}

void uvolni_pamet_pro_matici(void **x, void **p_x)
{
  free(*p_x);
  *p_x = NULL;
  free(*x);
  *x = NULL;
}

void zaloz_vektor_double(int n, double **x)
{
  if ((*x = (double *) malloc(n * sizeof(double))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
}

void uvolni_pamet_pro_vektor(void **x)
{
  free(*x);
  *x = NULL;
}

void nacti_data_ze_souboru(char *jmeno_souboru, int pocet_radku_dat, double **vstup_site, double **spravny_vystup_site, int vrstva1, int vrstva3)
{
  FILE *fr;
  int i, j, k;
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  for (i = 0; i < pocet_radku_dat; i++) {
    for (j = 0; j < vrstva1; j++) {
      if (fscanf(fr, "%lf", &vstup_site[i][j]) != 1) {
        printf("Chybi hodnota v souboru %s!\n", jmeno_souboru);
      }
    }
    for (k = 0; k < vrstva3; k++) {
      if (fscanf(fr, "%lf", &spravny_vystup_site[i][k]) != 1) {
        printf("Chybi hodnota v souboru %s!\n", jmeno_souboru);
      }
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
}

void inicializuj_matici_vah(double **w, int radek, int sloupek)
{
  int i, j;
  for (i = 0; i < radek; i++) {
    for (j = 0; j < sloupek; j++) {
      w[i][j] = real_rand();
    }
  }
}

void vypis_matici_vah(double **w, int radek, int sloupek, FILE *fw)
{
  char cislo_ret[DELKA_CISLA] = { '\0' };
  int i, j;
  for (i = 0; i < radek; i++) {
    for (j = 0; j < sloupek; j++) {
      ceske_cislo(cislo_ret, w[i][j]);
      fprintf(fw, "%s\t", cislo_ret);
    }
    fprintf(fw, "\n");
  }
}

void nacti_vahy(double **w, int radek, int sloupek, FILE *fr)
{
  int i, j;
  for (i = 0; i < radek; i++) {
    for (j = 0; j < sloupek; j++) {
      if (fread((void *) &(w[i][j]), sizeof(double), 1, fr) == 0) {
        printf("Nepodarila se nacist vaha!\n");
        exit(1);
      }
    }
  }
}

void zapis_vahy(double **w, int radek, int sloupek, FILE *fw)
{
  int i, j;
  for (i = 0; i < radek; i++) {
    for (j = 0; j < sloupek; j++) {
      if (fwrite((void *) &(w[i][j]), sizeof(double), 1, fw) != 1) {
        printf("Nepodarila se zapsat vaha!\n");
        exit(1);
      }
    }
  }
}

void nacti_nejlepsi(int *kroky_minule, NEJLEPSI *nejlepsi, FILE *fr)
{
  if (fread((void *) kroky_minule, sizeof(int), 1, fr) == 0) {
    printf("Nepodaril se nacist pocet pruchodu davkou dat!\n");
    exit(1);
  }
  if (fread((void *) nejlepsi, sizeof(NEJLEPSI), 1, fr) == 0) {
    printf("Nepodarila se nacist pocet pruchodu davkou dat!\n");
    exit(1);
  }
}

void zapis_nejlepsi(int kroky, NEJLEPSI *nejlepsi, char *jmeno_souboru)
{
  FILE *fw;
  if ((fw = fopen(jmeno_souboru, "wb")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  if (fwrite((void *) &kroky, sizeof(int), 1, fw) != 1) {
    printf("Nepodaril se zapsat pocet pruchodu davkou dat!\n");
    exit(1);
  }
  if (fwrite((void *) nejlepsi, sizeof(NEJLEPSI), 1, fw) != 1) {
    printf("Nepodaril se zapsat nejlepsi vysledek!\n");
    exit(1);
  }
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
}

int main()
{
  FILE *fr, *fw;
  char cislo_ret[DELKA_CISLA] = { '\0' };
  char cislo_ret_performance_tren[DELKA_CISLA] = { '\0' };
  int i, j, k, pocet_pruchodu_davkou_dat = 0;
  int vrstva1, vrstva2, vrstva3, vrstva4, vrstva5, pocet_radku_dat_tren, pocet_radku_dat_dev_test;
  int kroky_performance; /* pocet kroku po kterych nastane vypis performance */
  int kroky_konec; /* pocet kroku po kterych se ma skoncit */
  int kroky_minule; /* pocet kroku v minulych spustenich programu */
  int pocet_spravnych_vzoru_tren, pocet_spravnych_vzoru_dev_test, nejlepsi_pocet_spravnych_vzoru_dev_test;
  bool vystupni_neuron_souhlasi;
  double suma, stred_vystupu_neuronu, derivative_amplifier, parametr_rychlosti_uceni;
  double **vstup, *p_vstup, **vstup_dev_test, *p_vstup_dev_test;
  double **spravny_vystup, *p_spravny_vystup, **spravny_vystup_dev_test, *p_spravny_vystup_dev_test;
  double **w1, *p_w1/*, *vystup1, *chyba1*/;
  double **w2, *p_w2, *vystup2, *chyba2;
  double **w3, *p_w3, *vystup3, *chyba3;
  double **w4, *p_w4, *vystup4, *chyba4, *vystup5, *chyba5;
  double performance_tren, performance_dev_test;
  NEJLEPSI nejlepsi;
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER
  double strmost_sigmoidalni_funkce_hidden_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER)
  double minimum_sigmoidalni_funkce_hidden_layer, vyska_sigmoidalni_funkce_hidden_layer, strmost_sigmoidalni_funkce_hidden_layer;
#endif
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER
  double strmost_sigmoidalni_funkce_last_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER)
  double minimum_sigmoidalni_funkce_last_layer, vyska_sigmoidalni_funkce_last_layer, strmost_sigmoidalni_funkce_last_layer;
#endif
  unsigned int cas;
  if ((fr = fopen(SOUBOR_SETUP, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_SETUP);
    return 1;
  }
  vrstva1 = cti_int_ze_souboru(fr);
  vrstva2 = cti_int_ze_souboru(fr);
  vrstva3 = cti_int_ze_souboru(fr);
  vrstva4 = cti_int_ze_souboru(fr);
  vrstva5 = cti_int_ze_souboru(fr);
  kroky_performance = cti_int_ze_souboru(fr);
  kroky_konec = cti_int_ze_souboru(fr);
  parametr_rychlosti_uceni = cti_double_ze_souboru(fr);
  derivative_amplifier = cti_double_ze_souboru(fr);
  cas = cti_unsigned_int_ze_souboru(fr);
  if (cas == 0) {
    cas = time(NULL);
  }
  srand(cas);
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER
  strmost_sigmoidalni_funkce_hidden_layer = cti_double_ze_souboru(fr);
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER)
  minimum_sigmoidalni_funkce_hidden_layer = cti_double_ze_souboru(fr);
  vyska_sigmoidalni_funkce_hidden_layer = cti_double_ze_souboru(fr);
  strmost_sigmoidalni_funkce_hidden_layer = cti_double_ze_souboru(fr);
#endif
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER
  strmost_sigmoidalni_funkce_last_layer = cti_double_ze_souboru(fr);
  stred_vystupu_neuronu = 0.5;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER)
  minimum_sigmoidalni_funkce_last_layer = cti_double_ze_souboru(fr);
  vyska_sigmoidalni_funkce_last_layer = cti_double_ze_souboru(fr);
  strmost_sigmoidalni_funkce_last_layer = cti_double_ze_souboru(fr);
  stred_vystupu_neuronu = minimum_sigmoidalni_funkce_last_layer + vyska_sigmoidalni_funkce_last_layer / 2.0;
#else
  stred_vystupu_neuronu = 0.5;
#endif
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_SETUP);
    return 1;
  }
  pocet_radku_dat_tren = zjisti_pocet_radku_dat(SOUBOR_DATA_TRAINING);
  pocet_radku_dat_dev_test = zjisti_pocet_radku_dat(SOUBOR_DATA_DEV_TEST);
  zaloz_matici_double(pocet_radku_dat_tren, vrstva1, &vstup, &p_vstup);
  zaloz_matici_double(pocet_radku_dat_dev_test, vrstva1, &vstup_dev_test, &p_vstup_dev_test);
  zaloz_matici_double(pocet_radku_dat_tren, vrstva5, &spravny_vystup, &p_spravny_vystup);
  zaloz_matici_double(pocet_radku_dat_dev_test, vrstva5, &spravny_vystup_dev_test, &p_spravny_vystup_dev_test);
  zaloz_matici_double(vrstva1 + 1, vrstva2, &w1, &p_w1);
  zaloz_matici_double(vrstva2 + 1, vrstva3, &w2, &p_w2);
  zaloz_matici_double(vrstva3 + 1, vrstva4, &w3, &p_w3);
  zaloz_matici_double(vrstva4 + 1, vrstva5, &w4, &p_w4);
  zaloz_vektor_double(vrstva2, &vystup2);
  zaloz_vektor_double(vrstva3, &vystup3);
  zaloz_vektor_double(vrstva4, &vystup4);
  zaloz_vektor_double(vrstva5, &vystup5);
  zaloz_vektor_double(vrstva2, &chyba2);
  zaloz_vektor_double(vrstva3, &chyba3);
  zaloz_vektor_double(vrstva4, &chyba4);
  zaloz_vektor_double(vrstva5, &chyba5);
  nacti_data_ze_souboru(SOUBOR_DATA_TRAINING, pocet_radku_dat_tren, vstup, spravny_vystup, vrstva1, vrstva5);
  nacti_data_ze_souboru(SOUBOR_DATA_DEV_TEST, pocet_radku_dat_dev_test, vstup_dev_test, spravny_vystup_dev_test, vrstva1, vrstva5);
  if ((fr = fopen(SOUBOR_NEJLEPSI_BIN, "rb")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit. ", SOUBOR_NEJLEPSI_BIN);
    nejlepsi_pocet_spravnych_vzoru_dev_test = -1;
    nejlepsi.pocet_kroku = kroky_minule = pocet_pruchodu_davkou_dat;
    nejlepsi.performance_tren[0] = '\0';
    nejlepsi.pocet_spravnych_vzoru_tren = 0;
    nejlepsi.performance_dev_test[0] = '\0';
    nejlepsi.pocet_spravnych_vzoru_dev_test = 0;
    printf("Program bude vypisovat kroky od %d.\n", pocet_pruchodu_davkou_dat);
  } else {
      nacti_nejlepsi(&kroky_minule, &nejlepsi, fr);
      nejlepsi_pocet_spravnych_vzoru_dev_test = nejlepsi.pocet_spravnych_vzoru_dev_test;
      if (fclose(fr) == EOF) {
        printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_NEJLEPSI_BIN);
        return 1;
      }
    }
  if ((fr = fopen(SOUBOR_WEIGHTS_BIN, "rb")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit. ", SOUBOR_WEIGHTS_BIN);
    printf("Program se bude ucit z nahodnych vah.\n");
    inicializuj_matici_vah(w1, vrstva1 + 1, vrstva2);
    inicializuj_matici_vah(w2, vrstva2 + 1, vrstva3);
    inicializuj_matici_vah(w3, vrstva3 + 1, vrstva4);
    inicializuj_matici_vah(w4, vrstva4 + 1, vrstva5);
  } else {
    nacti_vahy(w1, vrstva1 + 1, vrstva2, fr);
    nacti_vahy(w2, vrstva2 + 1, vrstva3, fr);
    nacti_vahy(w3, vrstva3 + 1, vrstva4, fr);
    nacti_vahy(w4, vrstva4 + 1, vrstva5, fr);
    if (fclose(fr) == EOF) {
      printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_WEIGHTS_BIN);
      return 1;
    }
  }
  if ((fw = fopen(SOUBOR_PERFORMANCE, "a")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_PERFORMANCE);
    return 1;
  }
  fprintf(fw, "\n");
  kopiruj_soubor(SOUBOR_SETUP, fw);
  fprintf(fw, "\n%u\tSeed for random number generator\n\n", cas);
  while (pocet_pruchodu_davkou_dat < kroky_konec) {
    pocet_pruchodu_davkou_dat++;
    for (k = 0; k < pocet_radku_dat_tren; k++) {
      for (j = 0; j < vrstva2; j++) {
        suma = 0;
        for (i = 0; i < vrstva1; i++) {
          suma += vstup[k][i] * w1[i][j];
        }
        suma += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
        vystup2[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
        vystup2[j] = 1 / (1 + exp(-suma));
#endif
      }
      for (j = 0; j < vrstva3; j++) {
        suma = 0;
        for (i = 0; i < vrstva2; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
          suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup2[i]) * w2[i][j];
#else
          suma += vystup2[i] * w2[i][j];
#endif
        }
        suma += PRAH * w2[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
        vystup3[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
        vystup3[j] = 1 / (1 + exp(-suma));
#endif
      }
      for (j = 0; j < vrstva4; j++) {
        suma = 0;
        for (i = 0; i < vrstva3; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
          suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup3[i]) * w3[i][j];
#else
          suma += vystup3[i] * w3[i][j];
#endif
        }
        suma += PRAH * w3[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
        vystup4[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
        vystup4[j] = 1 / (1 + exp(-suma));
#endif
      }
      for (j = 0; j < vrstva5; j++) {
        suma = 0;
        for (i = 0; i < vrstva4; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
          suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup4[i]) * w4[i][j];
#else
          suma += vystup4[i] * w4[i][j];
#endif
        }
        suma += PRAH * w4[i][j];
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER
        vystup5[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
        chyba5[j] = spravny_vystup[k][j] - minimum_sigmoidalni_funkce_last_layer - vyska_sigmoidalni_funkce_last_layer * vystup5[j];
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER)
        vystup5[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
        chyba5[j] = spravny_vystup[k][j] - vystup5[j];
#else
        vystup5[j] = 1 / (1 + exp(-suma));
        chyba5[j] = spravny_vystup[k][j] - vystup5[j];
#endif
      }
      for (i = 0; i < vrstva4; i++) {
        chyba4[i] = 0;
        for (j = 0; j < vrstva5; j++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER
          chyba4[i] += derivative_amplifier * w4[i][j] * vystup5[j] * (1 - vystup5[j]) * chyba5[j] * vyska_sigmoidalni_funkce_last_layer * strmost_sigmoidalni_funkce_last_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER)
          chyba4[i] += derivative_amplifier * w4[i][j] * vystup5[j] * (1 - vystup5[j]) * chyba5[j] * strmost_sigmoidalni_funkce_last_layer;
#else
          chyba4[i] += derivative_amplifier * w4[i][j] * vystup5[j] * (1 - vystup5[j]) * chyba5[j];
#endif
        }
      }
      for (i = 0; i < vrstva3; i++) {
        chyba3[i] = 0;
        for (j = 0; j < vrstva4; j++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
          chyba3[i] += derivative_amplifier * w3[i][j] * vystup4[j] * (1 - vystup4[j]) * chyba4[j] * vyska_sigmoidalni_funkce_hidden_layer * strmost_sigmoidalni_funkce_hidden_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          chyba3[i] += derivative_amplifier * w3[i][j] * vystup4[j] * (1 - vystup4[j]) * chyba4[j] * strmost_sigmoidalni_funkce_hidden_layer;
#else
          chyba3[i] += derivative_amplifier * w3[i][j] * vystup4[j] * (1 - vystup4[j]) * chyba4[j];
#endif
        }
      }
      for (i = 0; i < vrstva2; i++) {
        chyba2[i] = 0;
        for (j = 0; j < vrstva3; j++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
          chyba2[i] += derivative_amplifier * w2[i][j] * vystup3[j] * (1 - vystup3[j]) * chyba3[j] * vyska_sigmoidalni_funkce_hidden_layer * strmost_sigmoidalni_funkce_hidden_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          chyba2[i] += derivative_amplifier * w2[i][j] * vystup3[j] * (1 - vystup3[j]) * chyba3[j] * strmost_sigmoidalni_funkce_hidden_layer;
#else
          chyba2[i] += derivative_amplifier * w2[i][j] * vystup3[j] * (1 - vystup3[j]) * chyba3[j];
#endif
        }
      }
      for (j = 0; j < vrstva2; j++) {
        for (i = 0; i < vrstva1; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
          w1[i][j] += parametr_rychlosti_uceni * vstup[k][i] * vystup2[j] * (1 - vystup2[j]) * chyba2[j] * vyska_sigmoidalni_funkce_hidden_layer * strmost_sigmoidalni_funkce_hidden_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          w1[i][j] += parametr_rychlosti_uceni * vstup[k][i] * vystup2[j] * (1 - vystup2[j]) * chyba2[j] * strmost_sigmoidalni_funkce_hidden_layer;
#else
          w1[i][j] += parametr_rychlosti_uceni * vstup[k][i] * vystup2[j] * (1 - vystup2[j]) * chyba2[j];
#endif
        }
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
        w1[i][j] += parametr_rychlosti_uceni * PRAH * vystup2[j] * (1 - vystup2[j]) * chyba2[j] * vyska_sigmoidalni_funkce_hidden_layer * strmost_sigmoidalni_funkce_hidden_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
        w1[i][j] += parametr_rychlosti_uceni * PRAH * vystup2[j] * (1 - vystup2[j]) * chyba2[j] * strmost_sigmoidalni_funkce_hidden_layer;
#else
        w1[i][j] += parametr_rychlosti_uceni * PRAH * vystup2[j] * (1 - vystup2[j]) * chyba2[j];
#endif
      }
      for (j = 0; j < vrstva3; j++) {
        for (i = 0; i < vrstva2; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
          w2[i][j] += parametr_rychlosti_uceni * (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup2[i]) * vystup3[j] * (1 - vystup3[j]) * chyba3[j] * vyska_sigmoidalni_funkce_hidden_layer * strmost_sigmoidalni_funkce_hidden_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          w2[i][j] += parametr_rychlosti_uceni * vystup2[i] * vystup3[j] * (1 - vystup3[j]) * chyba3[j] * strmost_sigmoidalni_funkce_hidden_layer;
#else
          w2[i][j] += parametr_rychlosti_uceni * vystup2[i] * vystup3[j] * (1 - vystup3[j]) * chyba3[j];
#endif
        }
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
        w2[i][j] += parametr_rychlosti_uceni * PRAH * vystup3[j] * (1 - vystup3[j]) * chyba3[j] * vyska_sigmoidalni_funkce_hidden_layer * strmost_sigmoidalni_funkce_hidden_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
        w2[i][j] += parametr_rychlosti_uceni * PRAH * vystup3[j] * (1 - vystup3[j]) * chyba3[j] * strmost_sigmoidalni_funkce_hidden_layer;
#else
        w2[i][j] += parametr_rychlosti_uceni * PRAH * vystup3[j] * (1 - vystup3[j]) * chyba3[j];
#endif
      }
      for (j = 0; j < vrstva4; j++) {
        for (i = 0; i < vrstva3; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
          w3[i][j] += parametr_rychlosti_uceni * (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup3[i]) * vystup4[j] * (1 - vystup4[j]) * chyba4[j] * vyska_sigmoidalni_funkce_hidden_layer * strmost_sigmoidalni_funkce_hidden_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          w3[i][j] += parametr_rychlosti_uceni * vystup3[i] * vystup4[j] * (1 - vystup4[j]) * chyba4[j] * strmost_sigmoidalni_funkce_hidden_layer;
#else
          w3[i][j] += parametr_rychlosti_uceni * vystup3[i] * vystup4[j] * (1 - vystup4[j]) * chyba4[j];
#endif
        }
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
        w3[i][j] += parametr_rychlosti_uceni * PRAH * vystup4[j] * (1 - vystup4[j]) * chyba4[j] * vyska_sigmoidalni_funkce_hidden_layer * strmost_sigmoidalni_funkce_hidden_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
        w3[i][j] += parametr_rychlosti_uceni * PRAH * vystup4[j] * (1 - vystup4[j]) * chyba4[j] * strmost_sigmoidalni_funkce_hidden_layer;
#else
        w3[i][j] += parametr_rychlosti_uceni * PRAH * vystup4[j] * (1 - vystup4[j]) * chyba4[j];
#endif
      }
      for (j = 0; j < vrstva5; j++) {
        for (i = 0; i < vrstva4; i++) {
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) && defined (CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER)
          w4[i][j] += parametr_rychlosti_uceni * (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup4[i]) * vystup5[j] * (1 - vystup5[j]) * chyba5[j] * vyska_sigmoidalni_funkce_last_layer * strmost_sigmoidalni_funkce_last_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) && defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER)
          w4[i][j] += parametr_rychlosti_uceni * (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup4[i]) * vystup5[j] * (1 - vystup5[j]) * chyba5[j] * strmost_sigmoidalni_funkce_last_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER)
          w4[i][j] += parametr_rychlosti_uceni * (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup4[i]) * vystup5[j] * (1 - vystup5[j]) * chyba5[j];
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER)
          w4[i][j] += parametr_rychlosti_uceni * vystup4[i] * vystup5[j] * (1 - vystup5[j]) * chyba5[j] * vyska_sigmoidalni_funkce_last_layer * strmost_sigmoidalni_funkce_last_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER)
          w4[i][j] += parametr_rychlosti_uceni * vystup4[i] * vystup5[j] * (1 - vystup5[j]) * chyba5[j] * strmost_sigmoidalni_funkce_last_layer;
#else
          w4[i][j] += parametr_rychlosti_uceni * vystup4[i] * vystup5[j] * (1 - vystup5[j]) * chyba5[j];
#endif
        }
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER
        w4[i][j] += parametr_rychlosti_uceni * PRAH * vystup5[j] * (1 - vystup5[j]) * chyba5[j] * vyska_sigmoidalni_funkce_last_layer * strmost_sigmoidalni_funkce_last_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER)
        w4[i][j] += parametr_rychlosti_uceni * PRAH * vystup5[j] * (1 - vystup5[j]) * chyba5[j] * strmost_sigmoidalni_funkce_last_layer;
#else
        w4[i][j] += parametr_rychlosti_uceni * PRAH * vystup5[j] * (1 - vystup5[j]) * chyba5[j];
#endif
      }
    }
    if (pocet_pruchodu_davkou_dat % kroky_performance == 0) {
      performance_tren = 0;
      pocet_spravnych_vzoru_tren = 0;
      for (k = 0; k < pocet_radku_dat_tren; k++) {
        for (j = 0; j < vrstva2; j++) {
          suma = 0;
          for (i = 0; i < vrstva1; i++) {
            suma += vstup[k][i] * w1[i][j];
          }
          suma += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          vystup2[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
          vystup2[j] = 1 / (1 + exp(-suma));
#endif
        }
        for (j = 0; j < vrstva3; j++) {
          suma = 0;
          for (i = 0; i < vrstva2; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
            suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup2[i]) * w2[i][j];
#else
            suma += vystup2[i] * w2[i][j];
#endif
          }
          suma += PRAH * w2[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          vystup3[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
          vystup3[j] = 1 / (1 + exp(-suma));
#endif
        }
        for (j = 0; j < vrstva4; j++) {
          suma = 0;
          for (i = 0; i < vrstva3; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
            suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup3[i]) * w3[i][j];
#else
            suma += vystup3[i] * w3[i][j];
#endif
          }
          suma += PRAH * w3[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          vystup4[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
          vystup4[j] = 1 / (1 + exp(-suma));
#endif
        }
        vystupni_neuron_souhlasi = true;
        for (j = 0; j < vrstva5; j++) {
          suma = 0;
          for (i = 0; i < vrstva4; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
            suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup4[i]) * w4[i][j];
#else
            suma += vystup4[i] * w4[i][j];
#endif
          }
          suma += PRAH * w4[i][j];
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER
          vystup5[j] = minimum_sigmoidalni_funkce_last_layer + vyska_sigmoidalni_funkce_last_layer / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER)
          vystup5[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
#else
          vystup5[j] = 1 / (1 + exp(-suma));
#endif
          chyba5[j] = spravny_vystup[k][j] - vystup5[j];
          performance_tren += chyba5[j] * chyba5[j];
          if ((vystup5[j] < stred_vystupu_neuronu && spravny_vystup[k][j] > stred_vystupu_neuronu) ||
              (vystup5[j] > stred_vystupu_neuronu && spravny_vystup[k][j] < stred_vystupu_neuronu)) {
            vystupni_neuron_souhlasi = false;
          }
        }
        if (vystupni_neuron_souhlasi) {
          pocet_spravnych_vzoru_tren++;
        }
      }
      fprintf(fw, "%d\t", pocet_pruchodu_davkou_dat + kroky_minule);
      ceske_cislo(cislo_ret_performance_tren, performance_tren);
      fprintf(fw, "%s\t", cislo_ret_performance_tren);
      fprintf(fw, "%d\t", pocet_spravnych_vzoru_tren);
      performance_dev_test = 0;
      pocet_spravnych_vzoru_dev_test = 0;
      for (k = 0; k < pocet_radku_dat_dev_test; k++) {
        for (j = 0; j < vrstva2; j++) {
          suma = 0;
          for (i = 0; i < vrstva1; i++) {
            suma += vstup_dev_test[k][i] * w1[i][j];
          }
          suma += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          vystup2[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
          vystup2[j] = 1 / (1 + exp(-suma));
#endif
        }
        for (j = 0; j < vrstva3; j++) {
          suma = 0;
          for (i = 0; i < vrstva2; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
            suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup2[i]) * w2[i][j];
#else
            suma += vystup2[i] * w2[i][j];
#endif
          }
          suma += PRAH * w2[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          vystup3[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
          vystup3[j] = 1 / (1 + exp(-suma));
#endif
        }
        for (j = 0; j < vrstva4; j++) {
          suma = 0;
          for (i = 0; i < vrstva3; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
            suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup3[i]) * w3[i][j];
#else
            suma += vystup3[i] * w3[i][j];
#endif
          }
          suma += PRAH * w3[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
          vystup4[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
          vystup4[j] = 1 / (1 + exp(-suma));
#endif
        }
        vystupni_neuron_souhlasi = true;
        for (j = 0; j < vrstva5; j++) {
          suma = 0;
          for (i = 0; i < vrstva4; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
            suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup4[i]) * w4[i][j];
#else
            suma += vystup4[i] * w4[i][j];
#endif
          }
          suma += PRAH * w4[i][j];
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER
          vystup5[j] = minimum_sigmoidalni_funkce_last_layer + vyska_sigmoidalni_funkce_last_layer / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER)
          vystup5[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
#else
          vystup5[j] = 1 / (1 + exp(-suma));
#endif
          chyba5[j] = spravny_vystup_dev_test[k][j] - vystup5[j];
          performance_dev_test += chyba5[j] * chyba5[j];
          if ((vystup5[j] < stred_vystupu_neuronu && spravny_vystup_dev_test[k][j] > stred_vystupu_neuronu) ||
              (vystup5[j] > stred_vystupu_neuronu && spravny_vystup_dev_test[k][j] < stred_vystupu_neuronu)) {
            vystupni_neuron_souhlasi = false;
          }
        }
        if (vystupni_neuron_souhlasi) {
          pocet_spravnych_vzoru_dev_test++;
        }
      }
      ceske_cislo(cislo_ret, performance_dev_test);
      fprintf(fw, "%s\t", cislo_ret);
      fprintf(fw, "%d\n", pocet_spravnych_vzoru_dev_test);
      fflush(fw);
      if (pocet_spravnych_vzoru_dev_test > nejlepsi_pocet_spravnych_vzoru_dev_test) {
        nejlepsi.pocet_kroku = pocet_pruchodu_davkou_dat + kroky_minule;
        nejlepsi_pocet_spravnych_vzoru_dev_test = pocet_spravnych_vzoru_dev_test;
        ceske_cislo(nejlepsi.performance_tren, performance_tren);
        nejlepsi.pocet_spravnych_vzoru_tren = pocet_spravnych_vzoru_tren;
        ceske_cislo(nejlepsi.performance_dev_test, performance_dev_test);
        nejlepsi.pocet_spravnych_vzoru_dev_test = pocet_spravnych_vzoru_dev_test;
      }
    }
  }
  fprintf(fw, "\n%d\t%s\t%d\t%s\t%d\t%s\n",
          nejlepsi.pocet_kroku, nejlepsi.performance_tren, nejlepsi.pocet_spravnych_vzoru_tren,
          nejlepsi.performance_dev_test, nejlepsi.pocet_spravnych_vzoru_dev_test, cislo_ret_performance_tren);
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_PERFORMANCE);
    return 1;
  }
  if ((fw = fopen(SOUBOR_VYSLEDEK, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_VYSLEDEK);
    return 1;
  }
  for (k = 0; k < pocet_radku_dat_tren; k++) {
    for (j = 0; j < vrstva2; j++) {
      suma = 0;
      for (i = 0; i < vrstva1; i++) {
        suma += vstup[k][i] * w1[i][j];
      }
      suma += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
      vystup2[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
      vystup2[j] = 1 / (1 + exp(-suma));
#endif
    }
    for (j = 0; j < vrstva3; j++) {
      suma = 0;
      for (i = 0; i < vrstva2; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
        suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup2[i]) * w2[i][j];
#else
        suma += vystup2[i] * w2[i][j];
#endif
      }
      suma += PRAH * w2[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
      vystup3[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
      vystup3[j] = 1 / (1 + exp(-suma));
#endif
    }
    for (j = 0; j < vrstva4; j++) {
      suma = 0;
      for (i = 0; i < vrstva3; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
        suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup3[i]) * w3[i][j];
#else
        suma += vystup3[i] * w3[i][j];
#endif
      }
      suma += PRAH * w3[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
      vystup4[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
      vystup4[j] = 1 / (1 + exp(-suma));
#endif
    }
    vystupni_neuron_souhlasi = true;
    for (j = 0; j < vrstva5; j++) {
      suma = 0;
      for (i = 0; i < vrstva4; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
        suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup4[i]) * w4[i][j];
#else
        suma += vystup4[i] * w4[i][j];
#endif
      }
      suma += PRAH * w4[i][j];
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER
      vystup5[j] = minimum_sigmoidalni_funkce_last_layer + vyska_sigmoidalni_funkce_last_layer / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER)
      vystup5[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
#else
      vystup5[j] = 1 / (1 + exp(-suma));
#endif
      ceske_cislo(cislo_ret, vystup5[j]);
      fprintf(fw, "%s\t", cislo_ret);
      if ((vystup5[j] < stred_vystupu_neuronu && spravny_vystup[k][j] > stred_vystupu_neuronu) ||
          (vystup5[j] > stred_vystupu_neuronu && spravny_vystup[k][j] < stred_vystupu_neuronu)) {
        vystupni_neuron_souhlasi = false;
      }
    }
    for (j = 0; j < vrstva5; j++) {
      fprintf(fw, "%g\t", spravny_vystup[k][j]);
    }
    if (vystupni_neuron_souhlasi) {
      fprintf(fw, "%s\n", "1");
    } else {
      fprintf(fw, "%s\n", "0");
    }
  }
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VYSLEDEK);
    return 1;
  }
  if ((fw = fopen(SOUBOR_VYSLEDEK_DEV_TEST, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_VYSLEDEK_DEV_TEST);
    return 1;
  }
  for (k = 0; k < pocet_radku_dat_dev_test; k++) {
    for (j = 0; j < vrstva2; j++) {
      suma = 0;
      for (i = 0; i < vrstva1; i++) {
        suma += vstup_dev_test[k][i] * w1[i][j];
      }
      suma += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
      vystup2[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
      vystup2[j] = 1 / (1 + exp(-suma));
#endif
    }
    for (j = 0; j < vrstva3; j++) {
      suma = 0;
      for (i = 0; i < vrstva2; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
        suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup2[i]) * w2[i][j];
#else
        suma += vystup2[i] * w2[i][j];
#endif
      }
      suma += PRAH * w2[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
      vystup3[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
      vystup3[j] = 1 / (1 + exp(-suma));
#endif
    }
    for (j = 0; j < vrstva4; j++) {
      suma = 0;
      for (i = 0; i < vrstva3; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
        suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup3[i]) * w3[i][j];
#else
        suma += vystup3[i] * w3[i][j];
#endif
      }
      suma += PRAH * w3[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_HIDDEN_LAYER)
      vystup4[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_hidden_layer));
#else
      vystup4[j] = 1 / (1 + exp(-suma));
#endif
    }
    vystupni_neuron_souhlasi = true;
    for (j = 0; j < vrstva5; j++) {
      suma = 0;
      for (i = 0; i < vrstva4; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_HIDDEN_LAYER
        suma += (minimum_sigmoidalni_funkce_hidden_layer + vyska_sigmoidalni_funkce_hidden_layer * vystup4[i]) * w4[i][j];
#else
        suma += vystup4[i] * w4[i][j];
#endif
      }
      suma += PRAH * w4[i][j];
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_LAST_LAYER
      vystup5[j] = minimum_sigmoidalni_funkce_last_layer + vyska_sigmoidalni_funkce_last_layer / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_LAST_LAYER)
      vystup5[j] = 1 / (1 + exp(-suma * strmost_sigmoidalni_funkce_last_layer));
#else
      vystup5[j] = 1 / (1 + exp(-suma));
#endif
      ceske_cislo(cislo_ret, vystup5[j]);
      fprintf(fw, "%s\t", cislo_ret);
      if ((vystup5[j] < stred_vystupu_neuronu && spravny_vystup_dev_test[k][j] > stred_vystupu_neuronu) ||
          (vystup5[j] > stred_vystupu_neuronu && spravny_vystup_dev_test[k][j] < stred_vystupu_neuronu)) {
        vystupni_neuron_souhlasi = false;
      }
    }
    for (j = 0; j < vrstva5; j++) {
      fprintf(fw, "%g\t", spravny_vystup_dev_test[k][j]);
    }
    if (vystupni_neuron_souhlasi) {
      fprintf(fw, "%s\n", "1");
    } else {
      fprintf(fw, "%s\n", "0");
    }
  }
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VYSLEDEK_DEV_TEST);
    return 1;
  }
  if ((fw = fopen(SOUBOR_VAHY, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_VAHY);
    return 1;
  }
  vypis_matici_vah(w1, vrstva1 + 1, vrstva2, fw);
  vypis_matici_vah(w2, vrstva2 + 1, vrstva3, fw);
  vypis_matici_vah(w3, vrstva3 + 1, vrstva4, fw);
  vypis_matici_vah(w4, vrstva4 + 1, vrstva5, fw);
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VAHY);
    return 1;
  }
  if ((fw = fopen(SOUBOR_WEIGHTS_BIN, "wb")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_WEIGHTS_BIN);
    return 1;
  }
  zapis_vahy(w1, vrstva1 + 1, vrstva2, fw);
  zapis_vahy(w2, vrstva2 + 1, vrstva3, fw);
  zapis_vahy(w3, vrstva3 + 1, vrstva4, fw);
  zapis_vahy(w4, vrstva4 + 1, vrstva5, fw);
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_WEIGHTS_BIN);
    return 1;
  }
  zapis_nejlepsi(pocet_pruchodu_davkou_dat + kroky_minule, &nejlepsi, SOUBOR_NEJLEPSI_BIN);
  uvolni_pamet_pro_matici((void **) &vstup, (void **) &p_vstup);
  uvolni_pamet_pro_matici((void **) &vstup_dev_test, (void **) &p_vstup_dev_test);
  uvolni_pamet_pro_matici((void **) &spravny_vystup, (void **) &p_spravny_vystup);
  uvolni_pamet_pro_matici((void **) &spravny_vystup_dev_test, (void **) &p_spravny_vystup_dev_test);
  uvolni_pamet_pro_matici((void **) &w1, (void **) &p_w1);
  uvolni_pamet_pro_matici((void **) &w2, (void **) &p_w2);
  uvolni_pamet_pro_matici((void **) &w3, (void **) &p_w3);
  uvolni_pamet_pro_matici((void **) &w4, (void **) &p_w4);
  uvolni_pamet_pro_vektor((void **) &vystup2);
  uvolni_pamet_pro_vektor((void **) &vystup3);
  uvolni_pamet_pro_vektor((void **) &vystup4);
  uvolni_pamet_pro_vektor((void **) &vystup5);
  uvolni_pamet_pro_vektor((void **) &chyba2);
  uvolni_pamet_pro_vektor((void **) &chyba3);
  uvolni_pamet_pro_vektor((void **) &chyba4);
  uvolni_pamet_pro_vektor((void **) &chyba5);
  return 0;
}
