#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <float.h>
#include <stdbool.h>
#include <locale.h>
#define DELKA_BUFFERU 16384
#define POCET_PLATNYCH_CISLIC 20
#define PRAH -1
#define SOUBOR_SETUP "setup.txt"
#define SOUBOR_DATA_TRAINING "DataTraining.txt"
#define SOUBOR_DATA_DEV_TEST "DataDevTest.txt"
#define SOUBOR_PERFORMANCE "performance.txt"
#define SOUBOR_VYSLEDEK "vysledek.txt"
#define SOUBOR_VYSLEDEK_DEV_TEST "vysledek_test.txt"
#define SOUBOR_VAHY "vahy.txt"
#define real_rand() ((double) rand() / (RAND_MAX + 1.0) - 0.5)
//#define CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER
#define CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER
#define CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER
//#define CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER

typedef struct {
  double performance_tren;
  int pocet_spravnych_vzoru_tren;
  double performance_dev_test;
  int pocet_spravnych_vzoru_dev_test;
  int pocet_kroku;
} NEJLEPSI;

void print_real_number(FILE *fw, double number)
{
  char retezec[DELKA_BUFFERU];
  char *podretezec;
  int rad;
  sprintf(retezec, "%e", number);
  podretezec = strchr(retezec, 'e');
  if (podretezec != NULL) {
    podretezec = strchr(podretezec, '-');
    if (podretezec != NULL) {
      sscanf(++podretezec, "%d", &rad);
      fprintf(fw, "%.*f", POCET_PLATNYCH_CISLIC + rad, number);
    } else {
      podretezec = strchr(retezec, '+');
      if (podretezec != NULL) {
        sscanf(++podretezec, "%d", &rad);
        fprintf(fw, "%.*f", POCET_PLATNYCH_CISLIC <= rad ? 0 : POCET_PLATNYCH_CISLIC - rad, number);
      } else {
        fprintf(fw, "%s\n", "Real number containing 'e' without any sign!");
      }
    }
  } else fprintf(fw, "%s\n", "Real number without 'e'!");
}

unsigned cti_unsigned_int_ze_souboru(FILE *fr)
{
  int c;
  unsigned cislo;
  if (fscanf(fr, "%u", &cislo) != 1)
    printf ("Chybi hodnota v souboru!\n");
  printf("Precteno %u", cislo);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
  return cislo;
}

int cti_int_ze_souboru(FILE *fr)
{
  int c;
  int cislo;
  if (fscanf(fr, "%d", &cislo) != 1)
    printf ("Chybi hodnota v souboru!\n");
  printf("Precteno %d", cislo);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
  return cislo;
}

double cti_double_ze_souboru(FILE *fr)
{
  int c;
  double cislo;
  if (fscanf(fr, "%lf", &cislo) != 1)
    printf ("Chybi hodnota v souboru!\n");
  printf("Precteno %g", cislo);
  while (((c = getc(fr)) != '\n') && !feof(fr))
    putchar(c);
  putchar('\n');
  return cislo;
}

void kopiruj_soubor(char *jmeno_souboru, FILE *fw)
{
  FILE *fr;
  int c;
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  while ((c = getc(fr)) != EOF)
    putc(c, fw);
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
}

int zjisti_pocet_radku_dat(char *jmeno_souboru)
{
  FILE *fr;
  char buffer[DELKA_BUFFERU];
  int pocet_radku = 0;
  double cislo;
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  while (fgets(buffer, DELKA_BUFFERU, fr)) {
    if (sscanf(buffer, "%lf", &cislo) == 1) {
      pocet_radku++;
    }
    buffer[0] = '\0';
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
  return pocet_radku;
}

void zaloz_matici_double(int radek, int sloupek, double ***x, double **p_x)
{
  int i;
  if ((*p_x = (double *) malloc(radek * sloupek * sizeof(double))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  if ((*x = (double **) malloc(radek * sizeof(double *))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  for (i = 0; i < radek; i++) {
    (*x)[i] = *p_x + i * sloupek;
  }
}

void uvolni_pamet_pro_matici_double(double ***x, double **p_x)
{
  free((void *) *p_x);
  *p_x = NULL;
  free((void *) *x);
  *x = NULL;
}

void zaloz_vektor_double(int n, double **x)
{
  if ((*x = (double *) malloc(n * sizeof(double))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
}

void uvolni_pamet_pro_vektor_double(double **x)
{
  free((void *) *x);
  *x = NULL;
}

void nacti_data_ze_souboru(char *jmeno_souboru, int pocet_radku_dat, double **vstup_site, double **spravny_vystup_site, int vrstva1, int vrstva3)
{
  FILE *fr;
  int i, j, k;
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  for (i = 0; i < pocet_radku_dat; i++) {
    for (j = 0; j < vrstva1; j++) {
      if (fscanf(fr, "%lf", &vstup_site[i][j]) != 1) {
        printf("Chybi hodnota v souboru %s!\n", jmeno_souboru);
      }
    }
    for (k = 0; k < vrstva3; k++) {
      if (fscanf(fr, "%lf", &spravny_vystup_site[i][k]) != 1) {
        printf("Chybi hodnota v souboru %s!\n", jmeno_souboru);
      }
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
}

void inicializuj_matici_vah(double **w, int radek, int sloupek)
{
  int i, j;
  for (i = 0; i < radek; i++) {
    for (j = 0; j < sloupek; j++) {
      w[i][j] = real_rand();
    }
  }
}

void vypis_matici_vah(double **w, int radek, int sloupek, FILE *fw)
{
  int i, j;
  for (i = 0; i < radek; i++) {
    for (j = 0; j < sloupek; j++) {
      print_real_number(fw, w[i][j]);
      fprintf(fw, "\t");
    }
    fprintf(fw, "\n");
  }
}

int main()
{
  FILE *fr, *fw;
  int i, j, k, l, m;
  int vrstva1, vrstva2, vrstva3, pocet_radku_dat_tren, pocet_radku_dat_dev_test;
  int kroky_performance; /* pocet kroku po kterych nastane vypis performance */
  int kroky_konec; /* pocet kroku po kterych se ma skoncit */
  int pocet_spravnych_vzoru_tren, pocet_spravnych_vzoru_dev_test, nejlepsi_pocet_spravnych_vzoru = -1;
  int t, suma_t = 0;
  bool vystupni_neuron_souhlasi;
  double stred_vystupu_neuronu, parametr_rychlosti_uceni;
  double **vstup, *p_vstup, **vstup_dev_test, *p_vstup_dev_test;
  double **spravny_vystup, *p_spravny_vystup, **spravny_vystup_dev_test, *p_spravny_vystup_dev_test;
  double **w1, *p_w1, *suma1, *vystup1, *chyba1;
  double **w2, *p_w2, *suma2, *vystup2, *chyba2;
  double *perform, performance_tren = 0.0, performance_dev_test, suma_perform, suma_t_perform, smernice = -1;
  NEJLEPSI nejlepsi;
  nejlepsi.performance_tren = 0.0;
  nejlepsi.pocet_spravnych_vzoru_tren = 0;
  nejlepsi.performance_dev_test = 0.0;
  nejlepsi.pocet_spravnych_vzoru_dev_test = 0;
  nejlepsi.pocet_kroku = 0;
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER
  double strmost_sigmoidalni_funkce_2nd_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER)
  double minimum_sigmoidalni_funkce_2nd_layer, vyska_sigmoidalni_funkce_2nd_layer, strmost_sigmoidalni_funkce_2nd_layer;
#endif
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER
  double strmost_sigmoidalni_funkce_3rd_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER)
  double minimum_sigmoidalni_funkce_3rd_layer, vyska_sigmoidalni_funkce_3rd_layer, strmost_sigmoidalni_funkce_3rd_layer;
#endif
  unsigned int cas;
  setlocale(LC_ALL,""); //https://www.cplusplus.com/reference/clocale/setlocale/
  if ((fr = fopen(SOUBOR_SETUP, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_SETUP);
    return 1;
  }
  vrstva1 = cti_int_ze_souboru(fr);
  vrstva2 = cti_int_ze_souboru(fr);
  vrstva3 = cti_int_ze_souboru(fr);
  kroky_performance = cti_int_ze_souboru(fr);
  kroky_konec = cti_int_ze_souboru(fr);
  parametr_rychlosti_uceni = cti_double_ze_souboru(fr);
  t = cti_int_ze_souboru(fr);
  cas = cti_unsigned_int_ze_souboru(fr);
  if (cas == 0) {
    cas = time(NULL);
  }
  srand(cas);
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER
  strmost_sigmoidalni_funkce_2nd_layer = cti_double_ze_souboru(fr);
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER)
  minimum_sigmoidalni_funkce_2nd_layer = cti_double_ze_souboru(fr);
  vyska_sigmoidalni_funkce_2nd_layer = cti_double_ze_souboru(fr);
  strmost_sigmoidalni_funkce_2nd_layer = cti_double_ze_souboru(fr);
#endif
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER
  strmost_sigmoidalni_funkce_3rd_layer = cti_double_ze_souboru(fr);
  stred_vystupu_neuronu = 0.5;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER)
  minimum_sigmoidalni_funkce_3rd_layer = cti_double_ze_souboru(fr);
  vyska_sigmoidalni_funkce_3rd_layer = cti_double_ze_souboru(fr);
  strmost_sigmoidalni_funkce_3rd_layer = cti_double_ze_souboru(fr);
  stred_vystupu_neuronu = minimum_sigmoidalni_funkce_3rd_layer + vyska_sigmoidalni_funkce_3rd_layer / 2.0;
#else
  stred_vystupu_neuronu = 0.5;
#endif
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_SETUP);
    return 1;
  }
  pocet_radku_dat_tren = zjisti_pocet_radku_dat(SOUBOR_DATA_TRAINING);
  pocet_radku_dat_dev_test = zjisti_pocet_radku_dat(SOUBOR_DATA_DEV_TEST);
  zaloz_matici_double(pocet_radku_dat_tren, vrstva1, &vstup, &p_vstup);
  zaloz_matici_double(pocet_radku_dat_dev_test, vrstva1, &vstup_dev_test, &p_vstup_dev_test);
  zaloz_matici_double(pocet_radku_dat_tren, vrstva3, &spravny_vystup, &p_spravny_vystup);
  zaloz_matici_double(pocet_radku_dat_dev_test, vrstva3, &spravny_vystup_dev_test, &p_spravny_vystup_dev_test);
  zaloz_matici_double(vrstva1 + 1, vrstva2, &w1, &p_w1);
  zaloz_matici_double(vrstva2 + 1, vrstva3, &w2, &p_w2);
  zaloz_vektor_double(vrstva2, &suma1);
  zaloz_vektor_double(vrstva3, &suma2);
  zaloz_vektor_double(vrstva2, &vystup1);
  zaloz_vektor_double(vrstva3, &vystup2);
  zaloz_vektor_double(vrstva2, &chyba1);
  zaloz_vektor_double(vrstva3, &chyba2);
  zaloz_vektor_double(t, &perform);
  nacti_data_ze_souboru(SOUBOR_DATA_TRAINING, pocet_radku_dat_tren, vstup, spravny_vystup, vrstva1, vrstva3);
  nacti_data_ze_souboru(SOUBOR_DATA_DEV_TEST, pocet_radku_dat_dev_test, vstup_dev_test, spravny_vystup_dev_test, vrstva1, vrstva3);
  inicializuj_matici_vah(w1, vrstva1 + 1, vrstva2);
  inicializuj_matici_vah(w2, vrstva2 + 1, vrstva3);
  if ((fw = fopen(SOUBOR_PERFORMANCE, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_PERFORMANCE);
    return 1;
  }
  for (i = 0; i < t; i++) {
    suma_t += i + 1;
    perform[i] = FLT_MAX / t;
  }
  m = 0;
  while (smernice <= 0 && m < kroky_konec) {
    m++;
    for (l = 0; l < pocet_radku_dat_tren; l++) {
      for (j = 0; j < vrstva2; j++) {
        suma1[j] = 0;
        for (i = 0; i < vrstva1; i++) {
          suma1[j] += vstup[l][i] * w1[i][j];
        }
        suma1[j] += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER)
        vystup1[j] = 1 / (1 + exp(-suma1[j] * strmost_sigmoidalni_funkce_2nd_layer));
#else
        vystup1[j] = 1 / (1 + exp(-suma1[j]));
#endif
      }
      for (k = 0; k < vrstva3; k++) {
        suma2[k] = 0;
        for (j = 0; j < vrstva2; j++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER
          suma2[k] += (minimum_sigmoidalni_funkce_2nd_layer + vystup1[j] * vyska_sigmoidalni_funkce_2nd_layer) * w2[j][k];
#else
          suma2[k] += vystup1[j] * w2[j][k];
#endif
        }
        suma2[k] += PRAH * w2[j][k];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER)
        vystup2[k] = 1 / (1 + exp(-suma2[k] * strmost_sigmoidalni_funkce_3rd_layer));
#else
        vystup2[k] = 1 / (1 + exp(-suma2[k]));
#endif
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER
        chyba2[k] = spravny_vystup[l][k] - minimum_sigmoidalni_funkce_3rd_layer - vystup2[k] * vyska_sigmoidalni_funkce_3rd_layer;
#else
        chyba2[k] = spravny_vystup[l][k] - vystup2[k];
#endif
      }
      for (j = 0; j < vrstva2; j++) {
        chyba1[j] = 0;
        for (k = 0; k < vrstva3; k++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER
          chyba1[j] += w2[j][k] * vystup2[k] * (1 - vystup2[k]) * chyba2[k] * vyska_sigmoidalni_funkce_3rd_layer * strmost_sigmoidalni_funkce_3rd_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER)
          chyba1[j] += w2[j][k] * vystup2[k] * (1 - vystup2[k]) * chyba2[k] * strmost_sigmoidalni_funkce_3rd_layer;
#else
          chyba1[j] += w2[j][k] * vystup2[k] * (1 - vystup2[k]) * chyba2[k];
#endif
        }
      }
      for (j = 0; j < vrstva2; j++) {
        for (i = 0; i < vrstva1; i++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER
          w1[i][j] += parametr_rychlosti_uceni * vstup[l][i] * vystup1[j] * (1 - vystup1[j]) * chyba1[j] * vyska_sigmoidalni_funkce_2nd_layer * strmost_sigmoidalni_funkce_2nd_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER)
          w1[i][j] += parametr_rychlosti_uceni * vstup[l][i] * vystup1[j] * (1 - vystup1[j]) * chyba1[j] * strmost_sigmoidalni_funkce_2nd_layer;
#else
          w1[i][j] += parametr_rychlosti_uceni * vstup[l][i] * vystup1[j] * (1 - vystup1[j]) * chyba1[j];
#endif
        }
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER
        w1[i][j] += parametr_rychlosti_uceni * PRAH * vystup1[j] * (1 - vystup1[j]) * chyba1[j] * vyska_sigmoidalni_funkce_2nd_layer * strmost_sigmoidalni_funkce_2nd_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER)
        w1[i][j] += parametr_rychlosti_uceni * PRAH * vystup1[j] * (1 - vystup1[j]) * chyba1[j] * strmost_sigmoidalni_funkce_2nd_layer;
#else
        w1[i][j] += parametr_rychlosti_uceni * PRAH * vystup1[j] * (1 - vystup1[j]) * chyba1[j];
#endif
      }
      for (k = 0; k < vrstva3; k++) {
        for (j = 0; j < vrstva2; j++) {
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER) && defined (CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER)
          w2[j][k] += parametr_rychlosti_uceni * (minimum_sigmoidalni_funkce_2nd_layer + vystup1[j] * vyska_sigmoidalni_funkce_2nd_layer) * vystup2[k] * (1 - vystup2[k]) * chyba2[k] * vyska_sigmoidalni_funkce_3rd_layer * strmost_sigmoidalni_funkce_3rd_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER) && defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER)
          w2[j][k] += parametr_rychlosti_uceni * (minimum_sigmoidalni_funkce_2nd_layer + vystup1[j] * vyska_sigmoidalni_funkce_2nd_layer) * vystup2[k] * (1 - vystup2[k]) * chyba2[k] * strmost_sigmoidalni_funkce_3rd_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER)
          w2[j][k] += parametr_rychlosti_uceni * (minimum_sigmoidalni_funkce_2nd_layer + vystup1[j] * vyska_sigmoidalni_funkce_2nd_layer) * vystup2[k] * (1 - vystup2[k]) * chyba2[k];
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER)
          w2[j][k] += parametr_rychlosti_uceni * vystup1[j] * vystup2[k] * (1 - vystup2[k]) * chyba2[k] * vyska_sigmoidalni_funkce_3rd_layer * strmost_sigmoidalni_funkce_3rd_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER)
          w2[j][k] += parametr_rychlosti_uceni * vystup1[j] * vystup2[k] * (1 - vystup2[k]) * chyba2[k] * strmost_sigmoidalni_funkce_3rd_layer;
#else
          w2[j][k] += parametr_rychlosti_uceni * vystup1[j] * vystup2[k] * (1 - vystup2[k]) * chyba2[k];
#endif
        }
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER
        w2[j][k] += parametr_rychlosti_uceni * PRAH * vystup2[k] * (1 - vystup2[k]) * chyba2[k] * vyska_sigmoidalni_funkce_3rd_layer * strmost_sigmoidalni_funkce_3rd_layer;
#elif defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER)
        w2[j][k] += parametr_rychlosti_uceni * PRAH * vystup2[k] * (1 - vystup2[k]) * chyba2[k] * strmost_sigmoidalni_funkce_3rd_layer;
#else
        w2[j][k] += parametr_rychlosti_uceni * PRAH * vystup2[k] * (1 - vystup2[k]) * chyba2[k];
#endif
      }
    }
    if (m % kroky_performance == 0) {
      performance_tren = 0;
      pocet_spravnych_vzoru_tren = 0;
      for (l = 0; l < pocet_radku_dat_tren; l++) {
        for (j = 0; j < vrstva2; j++) {
          suma1[j] = 0;
          for (i = 0; i < vrstva1; i++) {
            suma1[j] += vstup[l][i] * w1[i][j];
          }
          suma1[j] += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER)
          vystup1[j] = 1 / (1 + exp(-suma1[j] * strmost_sigmoidalni_funkce_2nd_layer));
#else
          vystup1[j] = 1 / (1 + exp(-suma1[j]));
#endif
        }
        vystupni_neuron_souhlasi = true;
        for (k = 0; k < vrstva3; k++) {
          suma2[k] = 0;
          for (j = 0; j < vrstva2; j++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER
            suma2[k] += (minimum_sigmoidalni_funkce_2nd_layer + vystup1[j] * vyska_sigmoidalni_funkce_2nd_layer) * w2[j][k];
#else
            suma2[k] += vystup1[j] * w2[j][k];
#endif
          }
          suma2[k] += PRAH * w2[j][k];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER)
          vystup2[k] = 1 / (1 + exp(-suma2[k] * strmost_sigmoidalni_funkce_3rd_layer));
#else
          vystup2[k] = 1 / (1 + exp(-suma2[k]));
#endif
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER
          vystup2[k] = minimum_sigmoidalni_funkce_3rd_layer + vystup2[k] * vyska_sigmoidalni_funkce_3rd_layer;
#endif
          chyba2[k] = spravny_vystup[l][k] - vystup2[k];
          performance_tren += chyba2[k] * chyba2[k];
          if ((vystup2[k] < stred_vystupu_neuronu && spravny_vystup[l][k] > stred_vystupu_neuronu) ||
              (vystup2[k] > stred_vystupu_neuronu && spravny_vystup[l][k] < stred_vystupu_neuronu)) {
            vystupni_neuron_souhlasi = false;
          }
        }
        if (vystupni_neuron_souhlasi) {
          pocet_spravnych_vzoru_tren++;
        }
      }
      print_real_number(fw, performance_tren);
      fprintf(fw, "\t%d\t", pocet_spravnych_vzoru_tren);
      performance_dev_test = 0;
      pocet_spravnych_vzoru_dev_test = 0;
      for (l = 0; l < pocet_radku_dat_dev_test; l++) {
        for (j = 0; j < vrstva2; j++) {
          suma1[j] = 0;
          for (i = 0; i < vrstva1; i++) {
            suma1[j] += vstup_dev_test[l][i] * w1[i][j];
          }
          suma1[j] += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER)
          vystup1[j] = 1 / (1 + exp(-suma1[j] * strmost_sigmoidalni_funkce_2nd_layer));
#else
          vystup1[j] = 1 / (1 + exp(-suma1[j]));
#endif
        }
        vystupni_neuron_souhlasi = true;
        for (k = 0; k < vrstva3; k++) {
          suma2[k] = 0;
          for (j = 0; j < vrstva2; j++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER
            suma2[k] += (minimum_sigmoidalni_funkce_2nd_layer + vystup1[j] * vyska_sigmoidalni_funkce_2nd_layer) * w2[j][k];
#else
            suma2[k] += vystup1[j] * w2[j][k];
#endif
          }
          suma2[k] += PRAH * w2[j][k];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER)
          vystup2[k] = 1 / (1 + exp(-suma2[k] * strmost_sigmoidalni_funkce_3rd_layer));
#else
          vystup2[k] = 1 / (1 + exp(-suma2[k]));
#endif
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER
          vystup2[k] = minimum_sigmoidalni_funkce_3rd_layer + vystup2[k] * vyska_sigmoidalni_funkce_3rd_layer;
#endif
          chyba2[k] = spravny_vystup_dev_test[l][k] - vystup2[k];
          performance_dev_test += chyba2[k] * chyba2[k];
          if ((vystup2[k] < stred_vystupu_neuronu && spravny_vystup_dev_test[l][k] > stred_vystupu_neuronu) ||
              (vystup2[k] > stred_vystupu_neuronu && spravny_vystup_dev_test[l][k] < stred_vystupu_neuronu)) {
            vystupni_neuron_souhlasi = false;
          }
        }
        if (vystupni_neuron_souhlasi) {
          pocet_spravnych_vzoru_dev_test++;
        }
      }
      print_real_number(fw, performance_dev_test);
      fprintf(fw, "\t%d\n", pocet_spravnych_vzoru_dev_test);
      fflush(fw);
      if (pocet_spravnych_vzoru_dev_test > nejlepsi_pocet_spravnych_vzoru) {
        nejlepsi_pocet_spravnych_vzoru = pocet_spravnych_vzoru_dev_test;
        nejlepsi.performance_tren = performance_tren;
        nejlepsi.pocet_spravnych_vzoru_tren = pocet_spravnych_vzoru_tren;
        nejlepsi.performance_dev_test = performance_dev_test;
        nejlepsi.pocet_spravnych_vzoru_dev_test = pocet_spravnych_vzoru_dev_test;
        nejlepsi.pocet_kroku = m;
      }
      for (i = 1; i < t; i++) {
        perform[i - 1] = perform[i];
      }
      perform[t - 1] = performance_dev_test;
      suma_perform = 0;
      suma_t_perform = 0;
      for (i = 0; i < t; i++) {
        suma_perform += perform[i];
        suma_t_perform += (i + 1) * perform[i];
      }
      smernice = t * suma_t_perform - suma_t * suma_perform;
    }
  }
  fprintf(fw, "\n");
  print_real_number(fw, nejlepsi.performance_tren);
  fprintf(fw, "\t%d\t", nejlepsi.pocet_spravnych_vzoru_tren);
  print_real_number(fw, nejlepsi.performance_dev_test);
  fprintf(fw, "\t%d\t%d\t", nejlepsi.pocet_spravnych_vzoru_dev_test, nejlepsi.pocet_kroku);
  print_real_number(fw, performance_tren);
  fprintf(fw, "\n%u\tSeed for random number generator\n\n", cas);
  kopiruj_soubor(SOUBOR_SETUP, fw);
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_PERFORMANCE);
    return 1;
  }

  if ((fw = fopen(SOUBOR_VYSLEDEK, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_VYSLEDEK);
    return 1;
  }
  for (l = 0; l < pocet_radku_dat_tren; l++) {
    for (j = 0; j < vrstva2; j++) {
      suma1[j] = 0;
      for (i = 0; i < vrstva1; i++) {
        suma1[j] += vstup[l][i] * w1[i][j];
      }
      suma1[j] += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER)
      vystup1[j] = 1 / (1 + exp(-suma1[j] * strmost_sigmoidalni_funkce_2nd_layer));
#else
      vystup1[j] = 1 / (1 + exp(-suma1[j]));
#endif
    }
    vystupni_neuron_souhlasi = true;
    for (k = 0; k < vrstva3; k++) {
      suma2[k] = 0;
      for (j = 0; j < vrstva2; j++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER
        suma2[k] += (minimum_sigmoidalni_funkce_2nd_layer + vystup1[j] * vyska_sigmoidalni_funkce_2nd_layer) * w2[j][k];
#else
        suma2[k] += vystup1[j] * w2[j][k];
#endif
      }
      suma2[k] += PRAH * w2[j][k];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER)
      vystup2[k] = 1 / (1 + exp(-suma2[k] * strmost_sigmoidalni_funkce_3rd_layer));
#else
      vystup2[k] = 1 / (1 + exp(-suma2[k]));
#endif
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER
      vystup2[k] = minimum_sigmoidalni_funkce_3rd_layer + vystup2[k] * vyska_sigmoidalni_funkce_3rd_layer;
#endif
      print_real_number(fw, vystup2[k]);
      fprintf(fw, "\t");
      if ((vystup2[k] < stred_vystupu_neuronu && spravny_vystup[l][k] > stred_vystupu_neuronu) ||
          (vystup2[k] > stred_vystupu_neuronu && spravny_vystup[l][k] < stred_vystupu_neuronu)) {
        vystupni_neuron_souhlasi = false;
      }
    }
    for (k = 0; k < vrstva3; k++) {
      fprintf(fw, "%g\t", spravny_vystup[l][k]);
    }
    if (vystupni_neuron_souhlasi) {
      fprintf(fw, "%s\n", "1");
    } else {
      fprintf(fw, "%s\n", "0");
    }
  }
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VYSLEDEK);
    return 1;
  }

  if ((fw = fopen(SOUBOR_VYSLEDEK_DEV_TEST, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_VYSLEDEK_DEV_TEST);
    return 1;
  }
  for (l = 0; l < pocet_radku_dat_dev_test; l++) {
    for (j = 0; j < vrstva2; j++) {
      suma1[j] = 0;
      for (i = 0; i < vrstva1; i++) {
        suma1[j] += vstup_dev_test[l][i] * w1[i][j];
      }
      suma1[j] += PRAH * w1[i][j];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_2ND_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER)
      vystup1[j] = 1 / (1 + exp(-suma1[j] * strmost_sigmoidalni_funkce_2nd_layer));
#else
      vystup1[j] = 1 / (1 + exp(-suma1[j]));
#endif
    }
    vystupni_neuron_souhlasi = true;
    for (k = 0; k < vrstva3; k++) {
      suma2[k] = 0;
      for (j = 0; j < vrstva2; j++) {
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_2ND_LAYER
        suma2[k] += (minimum_sigmoidalni_funkce_2nd_layer + vystup1[j] * vyska_sigmoidalni_funkce_2nd_layer) * w2[j][k];
#else
        suma2[k] += vystup1[j] * w2[j][k];
#endif
      }
      suma2[k] += PRAH * w2[j][k];
#if defined (CUSTOM_SIGMOIDAL_FUNCTION_SLOPE_3RD_LAYER) || defined (CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER)
      vystup2[k] = 1 / (1 + exp(-suma2[k] * strmost_sigmoidalni_funkce_3rd_layer));
#else
      vystup2[k] = 1 / (1 + exp(-suma2[k]));
#endif
#ifdef CUSTOM_SIGMOIDAL_FUNCTION_3RD_LAYER
      vystup2[k] = minimum_sigmoidalni_funkce_3rd_layer + vystup2[k] * vyska_sigmoidalni_funkce_3rd_layer;
#endif
      print_real_number(fw, vystup2[k]);
      fprintf(fw, "\t");
      if ((vystup2[k] < stred_vystupu_neuronu && spravny_vystup_dev_test[l][k] > stred_vystupu_neuronu) ||
          (vystup2[k] > stred_vystupu_neuronu && spravny_vystup_dev_test[l][k] < stred_vystupu_neuronu)) {
        vystupni_neuron_souhlasi = false;
      }
    }
    for (k = 0; k < vrstva3; k++) {
      fprintf(fw, "%g\t", spravny_vystup_dev_test[l][k]);
    }
    if (vystupni_neuron_souhlasi) {
      fprintf(fw, "%s\n", "1");
    } else {
      fprintf(fw, "%s\n", "0");
    }
  }
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VYSLEDEK_DEV_TEST);
    return 1;
  }

  if ((fw = fopen(SOUBOR_VAHY, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", SOUBOR_VAHY);
    return 1;
  }
  vypis_matici_vah(w1, vrstva1 + 1, vrstva2, fw);
  vypis_matici_vah(w2, vrstva2 + 1, vrstva3, fw);
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", SOUBOR_VAHY);
    return 1;
  }

  uvolni_pamet_pro_matici_double(&vstup, &p_vstup);
  uvolni_pamet_pro_matici_double(&vstup_dev_test, &p_vstup_dev_test);
  uvolni_pamet_pro_matici_double(&spravny_vystup, &p_spravny_vystup);
  uvolni_pamet_pro_matici_double(&spravny_vystup_dev_test, &p_spravny_vystup_dev_test);
  uvolni_pamet_pro_matici_double(&w1, &p_w1);
  uvolni_pamet_pro_matici_double(&w2, &p_w2);
  uvolni_pamet_pro_vektor_double(&suma1);
  uvolni_pamet_pro_vektor_double(&suma2);
  uvolni_pamet_pro_vektor_double(&vystup1);
  uvolni_pamet_pro_vektor_double(&vystup2);
  uvolni_pamet_pro_vektor_double(&chyba1);
  uvolni_pamet_pro_vektor_double(&chyba2);
  uvolni_pamet_pro_vektor_double(&perform);
  return 0;
}
